"""
Code for applying mass-balance approach to point sources.
"""

from collections import defaultdict
from datetime import datetime
import glob
import os
import multiprocessing as mp
import shutil
import textwrap
import warnings

from matplotlib.dates import DateFormatter
import cartopy.feature as cfeature
import matplotlib.pyplot as plt
import netCDF4
import numpy as np
import pandas
import scipy.optimize
import xarray

import smartcarb
import dplume

from . import timeprofiles as tp
from . import gauss
from . import mass
from . import misc
from . import vis

OUTPUT_PATH = '/output/SMARTCARB/output/mass-balance/PointSources'

LON_EQ_TO_LETTER = {
    0: 'a',  401: 'b', 805: 'c', 1207: 'd', 1610: 'e', 2012: 'f'
}

def add_ratio(data):
    """
    Add CO2:NOx ratio, but don't compute the ratio's uncertainity because number of estimates
    are often too small and individual uncertainties to large to compute a correct uncertainty
    by error propagation.
    """
    data['true_CO2:NO2_emissions'] = data['true_CO2_emissions'] / data['true_NO2_emissions']
    data['est_CO2:NO2_emissions'] = data['est_CO2_emissions'] / data['est_NO2_emissions'] # Mt/kt
    data['est_CO2:NO2_emissions_std'] = xarray.full_like(data['est_CO2_emissions_std'], np.nan)
    
    # convert Mt/kt -> Mt/Mt
    data['true_CO2:NO2_emissions'] *= 1000.0
    data['est_CO2:NO2_emissions'] *= 1000.0
    data['est_CO2:NO2_emissions_std'] *= 1000.0

    return data


def add_is_okay(data, source, path):
    """
    Check for a csv file with a column with `overpass` times and a second
    column `is_okay` with true or false that is used to mask bad estimates
    based on visual inspection. If the file does not exists, nothing will
    be done.
    """
    filename = os.path.join(OUTPUT_PATH, path, 'ts_%s_is_okay.csv' % source)
    
    try:
        is_okay = pandas.read_csv(filename, index_col=0, parse_dates=True)
    except FileNotFoundError: 
        print('Could not mask bad data: %s' % filename)
        return data
        
    okay_times = is_okay.index[is_okay['is_okay'].values]

    data['is_okay'] = xarray.DataArray(np.zeros(data.overpass.size, bool), dims='overpass')

    for i, t in enumerate(data.overpass):
        if t in okay_times:
            data['is_okay'].values[i] = True
            
    is_okay = data['is_okay'].values
    
    
    data['est_CO2_emissions'].values[~is_okay] = np.nan 
    data['est_NO2_emissions'].values[~is_okay] = np.nan     
    data['est_CO2_emissions_std'].values[~is_okay] = np.nan 
    data['est_NO2_emissions_std'].values[~is_okay] = np.nan     
    
    return data



def read_timeseries(source, nsat='6', tracer='NO2', ssp='high', scaling=1.0,
                    add_extra_error=True):
    """
    Create time series of estimate CO2 and NOx emissions by reading all
    netCDF files with flux estimates for a given source.
    
    Parameters
    ----------
    source           name of source (e.g. "Boxberg")
    nsat             number of satellites or constellation id (e.g. "2c", "3a", "6")
    tracer           trace gas used for plume detection (CO2 or NO2)
    ssp              single sounding precision scenario for trace gas ("low" or "high")
    scaling          scaling applied to anthropogenic NO2 signals (default: 1.0)
    add_extra_error  add estimate of uncertainty from method and background
                     that is not included in the estimated error
    """
    dataset = xarray.Dataset()
    
    for tracer in ['CO2', 'NO2']:
        em_true = smartcarb.io.read_emissions(origin=source, time='raw',
                                              use_constant=False,
                                              tracer=tracer)
        if tracer == 'NO2':
            em_true *= 1e3 # Mt yr-1 -> kt yr-1
            em_true *= scaling
            units = 'kt yr-1'
        else:
            units = 'Mt yr-1'

        dataset['time'] = xarray.DataArray(em_true.index, dims='time')
        dataset['true_%s_emissions' % tracer] = xarray.DataArray(em_true.values, dims='time')
        
        dataset['true_%s_emissions' % tracer].attrs['units'] = units
        dataset['true_%s_emissions' % tracer].attrs['species'] = tracer
    
    
    ts = defaultdict(list)
    
    # get filenames with flux information
    path = '%s-%s-scaling-%.1f' % (tracer, ssp, scaling)
    pattern = os.path.join(OUTPUT_PATH, path, source, 'Sentinel_7_CO2_*_flux.nc')
    fnames = sorted(glob.glob(pattern))
    
    lon_eqs_in_constellation = smartcarb.misc.lon_eq_by_nsat(nsat)
    
    for fn in fnames:
        with xarray.open_dataset(fn) as d:
            _, orbit, time, lon_eq = smartcarb.io.process_filename(fn)
            orbit = smartcarb.io.filename2orbit(fn)
            
            if lon_eq not in lon_eqs_in_constellation:
                print('Skip', time, lon_eq)
                continue
            
            ts['overpass'].append(time)
            ts['orbit'].append(orbit)
            ts['lon_eq'].append(lon_eq)
            ts['est_CO2_emissions'].append(d.CO2_flux_fit.attrs['emissions'])
            ts['est_CO2_emissions_std'].append(d.CO2_flux_fit.attrs['emissions_std'])           
            ts['est_NO2_emissions'].append(d.NO2_flux_fit.attrs['emissions'])
            ts['est_NO2_emissions_std'].append(d.NO2_flux_fit.attrs['emissions_std'])              
            ts['est_decay_time'].append(d.NO2_flux_fit.attrs['decay_time'])
            ts['est_decay_time_std'].append(d.NO2_flux_fit.attrs['decay_time_std'])
            ts['angle_between_curve_and_wind'].append(d.attrs['angle_between_curve_and_wind'])
            ts['wind_speed'].append(d.NO2_flux_fit.attrs['wind_speed'])
            ts['detected_pixels_upstream'].append(d.is_plume.isel(along=0).sum())

    
    for key, values in ts.items():
        dataset[key] = xarray.DataArray(values, dims='overpass')
        
    for tracer in ['CO2', 'NO2']:

        # add error estimate for mass-balance approach and background
        if add_extra_error:
            slope, intercept = {
                'CO2': (0.21, 5.42),
                'NO2': (0.25, 2.26)
            }[tracer]
            Q = float(dataset['true_%s_emissions' % tracer].mean())
            extra = slope * Q + intercept
            dataset['est_%s_emissions_std' % tracer] = np.sqrt(dataset['est_%s_emissions_std' % tracer]**2 + extra**2)
            
            print('Add extra uncertainty: %.2f' % extra)
        
        bad_values = (dataset['est_%s_emissions' % tracer] > 1000.0)
        bad_values |= (dataset['est_%s_emissions_std' % tracer] > 100.0)
        
        bad_values |= dataset['angle_between_curve_and_wind'].values > 45.0
        bad_values |= dataset['detected_pixels_upstream'].values >= 5
        
        print('Mask %d bad estimates.' % np.sum(bad_values))
        dataset['est_%s_emissions' % tracer].values[bad_values] = np.nan        
        dataset['est_%s_emissions_std' % tracer].values[bad_values] = np.nan
        
        
    dataset = add_is_okay(dataset, source, path=path)    
    
    # add ratio
    dataset= add_ratio(dataset)
        
        
    return dataset


def compute_statistics(data, tracer, lon_eqs_in_constellation,
                       true_annual_mean=0.0, quiet=False):
    """
    Compute mean bias, scatter etc for time series of estimate emissions
    or emission ratios.
    """
    valids = np.isin(data.lon_eq, lon_eqs_in_constellation)
    true_values = data['true_%s_emissions' % tracer].sel(time=data.overpass).values[valids]
    est_values = data['est_%s_emissions' % tracer].values[valids]

    n = np.isfinite(est_values).sum()
     
    mb = np.nanmean(est_values - true_values)
    sd = np.nanstd(est_values - true_values)
    nmb = 100*np.nanmean(est_values - true_values) / true_annual_mean
    nsd = 100*np.nanstd(est_values - true_values) / true_annual_mean
    rmsd = np.sqrt(mb**2 + sd**2 / n)
    nrmsd = 100 * rmsd / true_annual_mean
        
    median = np.nanmedian(data['est_%s_emissions_std' % tracer].values[valids])
    nmedian = 100 * median / true_annual_mean
    
    if not quiet:
        print('MB: %.1f (%.0f%%), SD: %.1f (%.0f%%), RMSD: %.1f (%.0f%%), N: %d, Median; %.1f (%.0f%%)' %
              (mb, nmb, sd, nsd, rmsd, nrmsd, n, median, nmedian))
        
    return {
        'bias': float(mb),           'bias [%]': float(nmb),
        'scatter': float(sd),        'scatter [%]': float(nsd),
        'RMSD': float(rmsd),         'RMSD [%]': float(nrmsd),
        'median_std': float(median), 'median_std [%]': float(nmedian),
        'number_of_cases': float(n),
        'true_annual_mean': float(true_annual_mean)
    }


def fit_ratio(ratio, ratio_std, use_sigma=False):
    """
    Fit a constant CO2:NOX ratio and its uncertainty
    """  
    xdata = np.arange(ratio.size)
    
    if use_sigma:
        sigma = ratio_std.values
        absolute_sigma = True
        ydata = ratio
        valids = np.isfinite(ydata) & np.isfinite(sigma)       
    else:
        sigma = None
        absolute_sigma = False
        ydata = ratio
        valids = np.isfinite(ydata)    

    # constant function
    constant = lambda xdata, p: np.full(xdata.shape, p)
    
    p, cov = scipy.optimize.curve_fit(constant, xdata[valids], ydata[valids],
                                      sigma=sigma, absolute_sigma=absolute_sigma)
    
    if use_sigma:
        p_std = np.sqrt(cov[0,0])
    else:
        # compute uncertainty assuming moderate quality of fit
        residual = constant(xdata[valids], p) - ydata[valids]

        chi2 = np.sum(residual**2) / (ydata.size - p.size)
        p_std = np.sqrt(chi2)

    return p[0], p_std


def prepare_ts(data, nsat, n_knots=2):
    """
    Prepare time series for constellation incl. computing seasonal cycle and
    annual emissions, and comparing with true emissions at overpass.
    """
    lon_eqs_in_constellation = smartcarb.misc.lon_eq_by_nsat(nsat)
    mask = data.lon_eq.isin(lon_eqs_in_constellation)
    
    # create new dataset for given constellation
    data = data.where(mask, drop=True)
    data.attrs['constellation'] = nsat
    
    # fix that where adds extra dimension
    for name in data:
        if 'time' in data[name].dims:
            data[name] = data[name].mean('overpass')
    
    # fit cycles
    ratio, ratio_std = fit_ratio(data['est_CO2:NO2_emissions'].values, None,
                                 use_sigma=False)
    data['est_CO2:NO2_emissions'].attrs['mean_ratio'] = ratio
    data['est_CO2:NO2_emissions'].attrs['mean_ratio_std'] = ratio_std
    
    stats = compute_statistics(data, 'CO2:NO2', lon_eqs_in_constellation,
                               true_annual_mean=data['true_CO2:NO2_emissions'].mean(),
                               quiet=True
                              )
    data['est_CO2:NO2_emissions'].attrs.update(stats)
    
    for trace_gas in ['CO2', 'NO2']:
        tp.add_cycle(data, n_knots=n_knots, nsat=nsat, gamma=None,
                     varname='est_%s_emissions' % trace_gas,
                     cyclename='est_%s_cycle' % trace_gas,
                     use_error=True)    
        
        true_annual_emissions = data['true_%s_emissions' % trace_gas].mean()
        stats = compute_statistics(data, trace_gas, lon_eqs_in_constellation,
                                   true_annual_mean=true_annual_emissions,
                                   quiet=True)

        data['est_%s_emissions' % trace_gas].attrs.update(stats)
        
    return data



def plot_ts(data, ymin=0, ymax=100, n_knots=2, show_errors=False):

    
    lon_eqs_in_constellation = sorted(set(data['lon_eq'].values.astype(int)))
    
    fig, axes = plt.subplots(3,1, sharex=True, figsize=(6.97, 6.97))


    for ax, tracer in zip(axes, ['CO2', 'NO2', 'CO2:NO2']):
        
        if tracer == 'NO2':
            units = 'kt yr$^{-1}$'
        elif tracer in ['CO2', 'CO2b']:
            units = 'Mt yr$^{-1}$'
        else:
            units = 'ratio'
            
        if tracer == 'CO2:NO2':
            ax.plot(data['time'], data['true_CO2:NO2_emissions'], 'k-',
                    label='True ratio: $R_t$ = %.0f' % np.mean(data['true_CO2:NO2_emissions']))
        else:
            em_date, em_range = vis.extract_time_range(data['time'],
                                                       data['true_%s_emissions' % tracer], 10, 11)
            ax.plot(em_date, em_range.mean(0), 'k-',
                    label='True emissions: %.1f %s' % (em_range.mean(), units))

            
        if tracer in ['CO2']:
            ax.set_ylabel('CO$_2$ [%s]' % units)
        elif tracer == 'NO2':
            ax.set_ylabel('NO$_x$ [%s]' % units)
        elif tracer == 'CO2b':
            ax.set_ylabel('NO$_x$ $\cdot$ $R_e$ [%s]' % units)   
        else:
            ax.set_ylabel('CO$_2$:NO$_x$')

            
        for i, lon_eq in enumerate(lon_eqs_in_constellation, 1):

            letter = LON_EQ_TO_LETTER[lon_eq]

            sat = data['lon_eq'].astype(int) == lon_eq

            times = data['overpass'].loc[sat].values
            values = data['est_%s_emissions' % tracer].loc[sat].values

            # use or not use uncertainty estimated from fitted cycle
            if tracer == 'CO2:NO2':
                errors = data['est_CO2:NO2_emissions'].attrs['mean_ratio_std']
            else:
                errors = data['est_%s_emissions_std' % tracer].loc[sat].values

            n_sats = np.sum(np.isfinite(values))

            if show_errors:
                ax.errorbar(times, values, errors, ls='', marker='o',
                            label='#%s (%d plumes)' % (letter, n_sats))
            else:
                ax.errorbar(times, values, ls='', marker='o',
                            label='#%s (%d plumes)' % (letter, n_sats))

        
        # plot cycle
        if tracer == 'CO2:NO2':
            
            ratio = data['est_CO2:NO2_emissions'].attrs['mean_ratio']
            ratio_std = data['est_CO2:NO2_emissions'].attrs['mean_ratio_std']            
            
            ax.hlines([ratio], data['time'][0], data['time'][-1], colors='r',
                      label='Mean ratio: $R_e$ = %.0f$\pm$%.0f' % (ratio, ratio_std))
        else:
            ax.plot(data['time'], data['est_%s_cycle' % tracer], 'r-',
                    label='Estimated emissions: %.1f$\pm$%.1f %s' %
                    (data['est_%s_cycle' % tracer].attrs['annual_mean'],
                     data['est_%s_cycle' % tracer].attrs['annual_mean_std'],
                     units))
        
        if tracer in ['CO2', 'CO2b', 'NO2']:
            ax.set_ylim(min(ymin, np.min(data['est_%s_emissions' % tracer])),
                        max(ymax, np.max(data['est_%s_emissions' % tracer]))
                       )
        else:
            true_ratio = np.mean(data['true_CO2:NO2_emissions'])
            bottom, top = ax.get_ylim()
            
            ax.set_ylim(
                min(bottom, np.floor(true_ratio / 500) * 500 - 1000),
                max(top, np.ceil(true_ratio / 500) * 500 + 1000)
            )

        ax.grid(True)

        ax.xaxis.set_major_formatter(DateFormatter('%b'))
        ax.set_xlim(datetime(2015,1,1), datetime(2015,12,31))

        ax.legend(fontsize='x-small', ncol=3, loc=1)
        
    fig.align_ylabels(axes)
    plt.tight_layout(pad=0.05)
    ax.set_xlabel('2015')
    plt.tight_layout(pad=0.05)
    
    return data




def plot_time_profiles():
    fig, axes = plt.subplots(1,3, figsize=(1.5*6.97, 1.5*4.31/3))
    tp.plot_diurnal_cycle('Janschwalde', norm=True, ax=axes[0], add_legend=False, only_co2=True)
    tp.plot_weekly_cycle('Janschwalde', norm=True, ax=axes[1], add_legend=False, only_co2=True)
    tp.plot_seasonal_cycle('Janschwalde', norm=True, ax=axes[2], add_legend=False, only_co2=True, short_month=True)

    for i, ax in enumerate(axes):
        ax.set_ylabel('')
        ax.set_ylim(0.6, 1.4)
        ax.set_title('(%s)' % 'abc'[i])


    axes[0].set_ylabel('Time profile')

    plt.tight_layout(pad=0.1)

    plt.savefig(os.path.join(OUTPUT_PATH, 'snap1_time_profiles.png'))
    plt.savefig(os.path.join(OUTPUT_PATH, 'snap1_time_profiles.pdf'))



def get_output_filename(filename, path='detections', extension='_detection.nc'):
    filename = os.path.splitext(os.path.basename(filename))[0]
    return os.path.join(OUTPUT_PATH, path, filename + extension)



def load_dataset(filename):
    
    with open(filename) as nc_file:
        data = xarray.open_dataset(filename).load()
        
    for key in data:
        if data[key].dtype == np.dtype('int8'):
            data[key] = data[key].astype(bool)
    
    return data



def write_to_netcdf(data, filename):
    """
    Write xarray.Dataset to netCDF. To avoid HDF5 errors in existing files,
    write a temporary file and replace existing file afterwards.
    """
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    
    # Write temporary file and close dataset.
    data.to_netcdf(filename + '.tmp')
    data.close()
    
    # Rename filename
    try:
        os.remove(filename)
    except FileNotFoundError:
        pass
    shutil.move(filename + '.tmp', filename)


    
def read_curves(filename, sources=None):
    """
    Read curves from netCDF file.
    """
    if sources is None:
        with netCDF4.Dataset(filename) as d:
            sources = list(d.groups.keys())
    
    curves = {}
        
    for source in sources:
        try:
            curves[source] = gauss.Poly2D.from_file(filename, group=source)
        except OSError as e:
            print(e.errno)
            
    return curves
    
    
def write_curves_to_netcdf(curves, filename):
    for source, curve in curves.items():
        if curve is not None:
            curve.to_file(filename=filename, group=source)
    

    
def select_source(data, source):
    """
    Select source in data but also compute new fields for other plumes and multiple sources.
    """
    this = data.sel(origin=source)
    this['other_sources'] = data['detected_plume'].any('origin') & ~this['detected_plume']
    this['multiple_sources'] = data['detected_plume'].sum('origin') > 1
    
    return this
    

def get_mass_balance_data(fn, origins, only_detection=False, make_figure=False,
                          redo_preperation=False, redo_detection=False, 
                          tracer='NO2', scaling=1.0, ssp='high', quiet=False):
    """
    
    """
    path = os.path.join('%s-%s-scaling-%.1f' % (tracer, ssp, scaling), 'detections')
    
    if tracer == 'NO2':
        co2_ssp = 0.7
        co2_scaling = 1.0
        
        no2_ssp = ssp
        no2_scaling = scaling
        
    elif tracer == 'CO2':
        co2_ssp = {'low': 0.5, 'medium': 0.7, 'high': 1.0}[ssp]
        co2_scaling = scaling
        
        no2_ssp = 'high'
        no2_scaling = 1.0
    
    
    output_fname = get_output_filename(fn, path=path, extension='_detection.nc')
    curves_fname = get_output_filename(fn, path=path, extension='_curves.nc')

    if not os.path.exists(output_fname) or redo_detection:
        if not quiet:
            print('Detect plumes')
        data = dplume.detect_plumes(fn, origins, tracer=tracer, ns=0.5, local_filter='gaussian',
                                    co2_ssp=co2_ssp, co2_scaling=co2_scaling,
                                    no2_ssp=no2_ssp, no2_scaling=no2_scaling)
        data.attrs.update(
            DESCRIPTION = ('Synthetic XCO2 and NO2 satellite image with ',
                           'auxiliary data for estimating CO2/NOx ',
                           'emissions of cities and point sources'),
            DATAORIGIN = 'SMARTCARB study',
            CREATOR = 'Gerrit Kuhlmann',
            EMAIL = 'gerrit.kuhlmann@empa.ch',
            AFFILIATION = 'Empa Duebendorf, Switzerland',
            DATE_CREATED = pandas.Timestamp.now().strftime('%Y-%m-%d %H:%M'),
        )
        
        write_to_netcdf(data, output_fname)
        
    if not quiet:    
        print('Load:', output_fname)
    data = load_dataset(output_fname)
    curves = {}
    
    
    if not only_detection:
        if redo_preperation or 'x' not in data:
        
            if not quiet:
                print('Prepare emissions') 
                
            data, curves = gauss.prepare_estimate_emissions(data, origin=None, plume_area=10e3)

            write_to_netcdf(data, output_fname)
            
            try:
                os.remove(curves_fname)
            except FileNotFoundError:
                pass
            
            write_curves_to_netcdf(curves, curves_fname)
            
            if not quiet:
                print('Load:', output_fname)
            data = load_dataset(output_fname)
            
            np.any(data['detected_plume']) 
            
        elif (np.any(data['detected_plume']) and any(not misc.has_multiple_sources(data, source)
                                                     for source in data.origin)):
            if not quiet:
                print('Read:', curves_fname)      
        
            curves = read_curves(curves_fname)
            
    
    if make_figure:
        for tracer in ['CO2', 'NO2']:
            fig = show_setup(data, curves, None, add_polygons=True, tracer=tracer)

            for ext in ['png', 'pdf']:
                figure_fname = get_output_filename(fn, path=path, extension='_%s_with_detections.%s' % (tracer, ext))
                fig.savefig(figure_fname)
    
    return data, curves



def show_setup(data, curves, ld=None, add_polygons=True, tracer='CO2', zoom_on=None, figwidth=6.97,
               ax=None, cax=None):
    """
    Show detected plumes in swath and add curves and control polygons if available.
    """
    first = True # first time adding polygons
    
    fig = dplume.vis.visualize(data, do_zoom=True, show_clouds=True, legend_loc='upper left',
                               tracer=tracer, zoom_on=zoom_on, figwidth=figwidth, ax=ax, cax=cax)
    
    if ax is None:
        ax, cax = fig.get_axes()
        

    for i, source in enumerate(data.origin.values):
    
        if zoom_on is not None and source != zoom_on:
            continue
    
        d = data.sel(origin=source)

        if source in curves and curves[source] is not None:
            
            plume = d['detected_plume']
        
            if ld is None:
                xa, xb, ya, yb = compute_polygons(d, delta=5e3,
                                                  add_upstream_box=True)
            else:
                xa, xb = ld[i].xa, ld[i].xb
                ya, yb = ld[i].ya, ld[i].yb
                
            
            # don't draw curves for overlapping plumes
            if misc.has_multiple_sources(data, source):
                continue
                
            x = d['x'].values[plume]
            y = d['y'].values[plume]
            curve = curves[source]
         
            tmax = curve.arc2parameter(xb[-1])
            t = np.linspace(curve.t_o, tmax)            
            
            smartcarb.map.add_curve(ax, *curve(t=t), origin=source, zorder=10,
                                    color='black', label='Center lines' if i==0 else None)
            
            if add_polygons:
                
                # polygon upstream
                mass.add_area(ax, d, curve, xa[:1], xb[:1], ya[:1], yb[:1],
                              origin=source, add_label=first, lw=1,
                              do_middle_lines=True)     
                first = False
                
                # polygons downstream
                mass.add_area(ax, d, curve, xa[1:], xb[1:], ya[1:], yb[1:],
                              origin=source, add_label=first, lw=1,
                              do_middle_lines=True)     
    return fig



def plot_mass_balance(ld, time, origin, figwidth=6.97, ax1=None, ax2=None, no2_scaling=1.0):

    if ax1 is None or ax2 is None:
        fig, (ax1, ax2) = plt.subplots(2,1, sharex=True, figsize=(6.97,4))
    else:
        fig = ax1.get_figure()

    factor = gauss.kgs_to_Mtyr(1.0)
    
    ax1.errorbar(ld['along'] / 1e3, factor * ld['CO2_flux'], factor * ld['CO2_flux_std'],
                 marker='o', ms=4, ls='', label='along-plume estimates')
    ax2.errorbar(ld['along'] / 1e3, 1e3 * factor * ld['NO2_flux'], 1e3 * factor * ld['NO2_flux_std'],
                 marker='o', ms=4, ls='', label='along-plume estimates')

    
    # CO2
    Q = ld['CO2_flux_fit'].attrs['emissions']
    u = ld['CO2_flux_fit'].attrs['wind_speed']
    Q_std = ld['CO2_flux_fit'].attrs['emissions_std']

    x = np.linspace(1.0, ld['along'][-1] + 2.5e3, 501)
    y = Q * np.ones_like(x)
    
    label = 'estimated (Q = %.1f$\pm$%.1f Mt yr$^{-1}$, u = %.1f m s$^{-1}$)' % (Q, Q_std, u)   
    ax1.plot(x, y, label=label, color='C0', ls='--')

    # NO2
    Q = ld['NO2_flux_fit'].attrs['emissions']
    u = ld['NO2_flux_fit'].attrs['wind_speed']
    tau = ld['NO2_flux_fit'].attrs['decay_time']
    
    Q_std = ld['NO2_flux_fit'].attrs['emissions_std']
    tau_std = ld['NO2_flux_fit'].attrs['decay_time_std']
   
    if np.isnan(tau):
        y = np.full(x.shape, Q)
    else:
        y = gauss.point_plume_model(x, Q, u * tau * 3600)
        
    label = 'estimated (Q = %.1f$\pm$%.1f kt yr$^{-1}$, u = %.1f m s$^{-1}$, $\\tau$ = %.1f$\pm$%.1f h)' % (Q, Q_std, u, tau, tau_std)    
    ax2.plot(x / 1e3, y, label=label, color='C0', ls='--')

    # true emissions
    q_CO2 = gauss.read_true_emissions(time, tracer='CO2', origin=origin, tmin=10, tmax=11).mean()
    q_NOx = no2_scaling * gauss.read_true_emissions(time, tracer='NO2', origin=origin, tmin=10, tmax=11).mean()   
    
    right = ld['along'][-1] / 1e3 + 2.5
    ax1.hlines([q_CO2], 0.0, right, color='C1', label='true (Q = %.1f Mt yr$^{-1}$)' % q_CO2)
    ax2.hlines([q_NOx], 0.0, right, color='C1', label='true (Q = %.1f kt yr$^{-1}$)' % q_NOx)

    # labels
    ax1.set_ylabel('CO$_2$ flux [Mt yr$^{-1}$]')
    ax2.set_ylabel('NO$_\mathrm{x}$ flux [kt yr$^{-1}$]')
    
    for ax in [ax1, ax2]:
        if q_NOx < 25:
            ax.set_ylim(-10, 60)        
        else:
            ax.set_ylim(-25, 100)
        ax.grid(True)

        ax.legend(fontsize='small')

        right = ld['along'].max() / 1e3 + 5.0
        xticks = np.array( sorted(set(list(ld.xa.values) + list(ld.xb.values))) ) / 1e3
        ax.set_xticks(xticks)
        ax.set_xlim(xticks[0], xticks[-1])
        
    ax2.set_xlabel('Along plume distance [km]')
    

    
    return fig



def add_lines(ax, edgecolor='k'):
    ax.coastlines(resolution='10m', color=edgecolor, linewidth=1.0)
    lines = cfeature.NaturalEarthFeature(
        category='cultural',
        name='admin_0_boundary_lines_land',
        scale='10m',
    )
    ax.add_feature(lines, edgecolor=edgecolor, facecolor='none', linewidth=1.0)

    lines = cfeature.NaturalEarthFeature(
            category='cultural',
            name='admin_1_states_provinces_lines',
            scale='10m',
    )
    ax.add_feature(lines, edgecolor=edgecolor, facecolor='none', linewidth=0.5)
    

    
def add_text(ax, data, ld, source):
    time = pandas.Timestamp(data.time.values)
    plume_length = ld.x.values[ld.is_plume.values].max() / 1e3
    plume_age = plume_length / (ld.CO2_flux_fit.attrs['wind_speed'] * 3600 / 1000)

    text = """\
        {satellite} ({time}, Orbit: {orbit})

        Instrument scenario:
        • CO2: $\sigma$ = 0.7 ppm
        • NO2: $\sigma$ = 2$\\times$10$^{{15}}$ molcules/cm²

        Plume information:
        • Detection: {detection_tracer} (q = {q}, $\sigma_n$ = {ns} px)
        • Plume size: {plume_size:d} px
        • Plume length: {plume_length:.1f} km
        • Estimated plume age: {plume_age:.1f} h
        • Mean wind [0..1 km]: {speed:.1f} m/s, {direction:.0f} degrees
        • Angle between curve and wind: {angle:.0f} degrees

        True emissions (10:30 UTC):
        • CO2: {true_CO2_flux:.1f} Mt/yr
        • NOx: {true_NO2_flux:.1f} kt/yr

        Estimated emissions:
        • CO2 flux: {CO2_flux:.1f}$\pm${CO2_flux_std:.1f} Mt/yr
        • NOx flux: {NO2_flux:.1f}$\pm${NO2_flux_std:.1f} kt/yr
        • NOx decay time: {NOX_decay:.1f}$\pm${NOX_decay_std:.1f} h


    """.format(satellite='CO2M%s' % vis.LON_EQ_TO_LETTER.get(data.lon_eq, 'x'),
               time=time.strftime('%d %b %Y, %H:%M UTC'),
               orbit=data.orbit,
               detection_tracer=data.attrs['tracer'].replace('2', '$_2$'),
               q=data.attrs['probability for z-value'],
               ns=data.attrs['size of neighborhood'],
               speed=ld.CO2_flux_fit.attrs['wind_speed'],
               direction=ld.CO2_flux_fit.attrs['wind_direction'],
               angle=ld.attrs['angle_between_curve_and_wind'],
               CO2_flux=ld.CO2_flux_fit.attrs['emissions'],           
               CO2_flux_std=ld.CO2_flux_fit.attrs['emissions_std'],  
               NO2_flux=ld.NO2_flux_fit.attrs['emissions'],           
               NO2_flux_std=ld.NO2_flux_fit.attrs['emissions_std'],       
               NOX_decay=ld.NO2_flux_fit.attrs['decay_time'],           
               NOX_decay_std=ld.NO2_flux_fit.attrs['decay_time_std'],
               plume_length=plume_length,
               plume_age=plume_age,
               plume_size=int(ld.is_plume.values.sum()),
               true_CO2_flux=gauss.read_true_emissions(time, 'CO2', source, tmin=10, tmax=11).mean(),
               true_NO2_flux=gauss.read_true_emissions(time, 'NO2', source, tmin=10, tmax=11).mean()
              )
    
    ax.text(0.02, 0.98, textwrap.dedent(text), va='top', transform=ax.transAxes)

    

def plot_source_overview(data, curves, line_densities, source, no2_scaling=1.0):
    domain = smartcarb.cfg.DOMAINS['smartcarb']
    time = pandas.Timestamp(data.time.values)

    with plt.style.context({'font.size': 8}):
        # figure size (13.94, 6.5) fits on wide screen power point and two figures per page in Copernicus journal.
        scaling = 1.0
        figsize= (13.94*scaling, 6.5 * scaling)
        fig = plt.figure(figsize=figsize)

        # size of level-2 image
        aspect = (smartcarb.cfg.rlon[-1] - smartcarb.cfg.rlon[0]) / (smartcarb.cfg.rlat[-1] - smartcarb.cfg.rlat[0])
        height = (figsize[1] - 0.7) / 2
        width =  height * aspect
        dy = 0.7 / 4 / figsize[1]
        rel_height = height / figsize[1]
        rel_width = width / figsize[0]

        axes = [None, None]
        caxes = [None, None]

        for i in [0,1]:
            axes[i] = fig.add_axes([
                0.01,
                dy + i * (rel_height + 2 * dy),
                rel_width,
                rel_height],
                projection=domain.proj
            )
            axes[i].set_aspect('equal', adjustable='box')
            axes[i].set_xlim(domain.startlon, domain.stoplon)
            axes[i].set_ylim(domain.startlat, domain.stoplat)
            add_lines(axes[i])

            axes[i].set_title('(%s)' % 'ca'[i])

            caxes[i] = fig.add_axes([
                0.02 + rel_width,
                dy + i * (rel_height + 2 * dy),
                0.01, 
                rel_height]
            )

            show_setup(data, curves, None, ax=axes[i], cax=caxes[i], 
                       zoom_on=source, tracer=['NO2', 'CO2'][i])

        # mass balance
        h = (dy + rel_height - 0.06 - 0.015) / 2
        left = 0.02 + rel_width + 0.12

        # other axes
        ax1 = fig.add_axes([
            left,
            0.06 + 0.015 + h,
            0.38,
            h,
        ])
        ax2 = fig.add_axes([
            left,
            0.06,
            0.38,
            h])
        
    
        plot_mass_balance(line_densities, time=time, origin=source, ax1=ax1,
                          ax2=ax2, no2_scaling=no2_scaling)       
        ax1.set_xticklabels([])
        ax1.set_title('(d)')


        w = 0.02 + rel_width + 0.12 + 0.38 + 0.02
        
        tax = fig.add_axes([
            w,
            0.01,
            1.0 - w - 0.02,     
            2*h + 0.015 + 0.05,
        ])
        
        tax.set_xticklabels([])
        tax.set_yticklabels([])
        add_text(tax, data, line_densities, source)
        
        
        # cross sections for first six cases (  0.98)
        axes = []
        
        
        # upper row
        for i in [0,1,2]:
            ax = fig.add_axes([
                left + i * (0.6/3 - 0.01 + 0.01),
                dy + (rel_height + 2 * dy) + 0.05 + h,
                0.6/3 - 0.01,
                h - 0.01,
            ])
            
            axes.append(ax)
        
        # lower row
        for i in [0,1,2]:            
            ax = fig.add_axes([
                left + i * (0.6/3 - 0.01 + 0.01),
                dy + (rel_height + 2 * dy) + 0.04,
                0.6/3 - 0.01,
                h - 0.01,
            ])
            axes.append(ax)
        
        # plot line densities
        n = line_densities.along.size
        
        if n <= 6:
            alongs = np.arange(n)
        else:
            alongs = [0, 1, 2, 3, n-2, n-1]
        
        for i, ax in zip(alongs, axes):
            
            if 0 <= i < line_densities.along.size:
                polygon = line_densities.isel(along=i)
                
                with warnings.catch_warnings():
                    warnings.filterwarnings('ignore', r'All-NaN (slice|axis) encountered')
                    
                    vis.plot_across_section(polygon, tracer='CO2',
                                            method=polygon.method,
                                            show_true=False,
                                            add_errors=True, ax=ax, no_twinx=True,
                                            legend='simple')

                ax.text(0.01, 0.98, '%d-%d km' % (polygon.xa/1e3, polygon.xb/1e3),
                        va='top', ha='left', transform=ax.transAxes)
            
        bottom = min(ax.get_ylim()[0] for ax in axes)
        top = max(ax.get_ylim()[1] for ax in axes)
        
        # update labels etc.
        for i, ax in enumerate(axes):
            ax.set_ylim(bottom, top)
            
            if i == 1:
                ax.set_title('(b)')
            
            if i in [0,1,2]:
                ax.set_xticklabels([])
                ax.set_xlabel('')
            
            if i not in [0,3]:
                ax.set_yticklabels([])
                ax.set_ylabel('')
                
            if i != 2:
                legend = ax.get_legend()
                
                if legend is not None:
                    legend.remove()

        
        return fig


def compute_polygons(data, delta=5e3, add_upstream_box=True):
    """
    Compute [xa,xb] and [ya,yb] intervals for polygons.
    """
    dmax = np.nanmax(data.xp.values[data.detected_plume])
    distances = np.arange(0.0, dmax+delta, delta)
   
    if add_upstream_box:
        xa_values = np.concatenate([[-12e3], distances[:-1]])
        xb_values = np.concatenate([[-2e3], distances[1:]])
    else:
        xa_values = distances[:-1]
        xb_values = distances[1:]
    
    ya, yb = mass.get_plume_width(data, dy=2e3)
    
    return xa_values, xb_values, np.full_like(xa_values, ya), np.full_like(xb_values, yb)
    
    
    

def estimate_fluxes(data, curve, method='two-gauss',
                    origin=None, show=False, delta=5e3):
    """
    Estimate fluxes in along-plume direction.
    """
    time = pandas.Timestamp(data['time'].values)
    
    wind_info = gauss.get_wind_speed(time, hmin=0.0, hmax=1000.0, origin=origin)
    wind_speed = wind_info['mean wind speed (m/s)']
    wind_direction = wind_info['mean wind direction (deg)']

    # add one polygon [-10 km - -2 km] upstream for checking for upstream sources
    xa_values, xb_values, ya, yb = compute_polygons(data, delta=delta, add_upstream_box=True)
    
    results = []
    
    for xa, xb in zip(xa_values, xb_values):
            
        if method == 'two-gauss':
            
            # use "sub-areas" upstream where no plume should exists 
            this_method = 'two-gauss' if xa >= 0.0 else 'sub-areas'
                
            ld = mass.extract_line_density(data, 'CO2_minus_estimated_background_mass', 
                                           xa=xa, xb=xb, ya=ya, yb=yb, show=show, show_true=False,
                                           method=this_method, share_mu=True)
            results.append(ld)
            
            
        else:
            raise NotImplementedError
            ld = mass.extract_line_density(data, 'CO2_minus_estimated_background_mass', 
                                           xa=xa, xb=xb, show=False, show_true=False,
                                           method=method)
            
            results['CO2_line_density'][i] = ld['sub-areas']
            results['CO2_line_density_std'][i] = ld['sub-areas_std']
            
            
            ld = mass.extract_line_density(data, 'NO2_minus_estimated_background_mass', 
                                           xa=xa, xb=xb, show=False, show_true=False,
                                           method=method)
            
            # NO2/NOX flux (using 1.32 conversion factor, Beirle et al. 2011)
            results['NO2_line_density'][i] = 1.32 * ld['sub-areas']
            results['NO2_line_density_std'][i] = 1.32 * ld['sub-areas_std']
   
    if len(results) == 0:
        return None
    
    results = mass.concat_polygons(results)
    results['along'] = xarray.DataArray(0.5 * (results.xa + results.xb), dims='along')
    
    for tracer in ['CO2', 'NO2']:
        
        wind_speed_std = 2.0
        
        results['%s_flux' % tracer] = wind_speed * results['%s_line_density' % tracer]
        results['%s_flux_std' % tracer] = np.sqrt(
            wind_speed**2 * results['%s_line_density_std' % tracer]**2 + 
            wind_speed_std**2 * results['%s_line_density' % tracer]**2
        )
        print('Flux: %2g' % np.mean(results['%s_flux_std' % tracer]))
        
        q, q_std, fit, p, p_std = gauss.fit_emissions(tracer,
                                                      results['along'].values,
                                                      results['%s_flux' % tracer].values,
                                                      results['%s_flux_std' % tracer].values,
                                                      dmin_fit=0.0, dmax_fit=np.inf)
        

        angle = gauss.compute_angle_between_curve_and_wind(origin, data, curve, hmin=0.0, hmax=1e3)
        results.attrs['angle_between_curve_and_wind'] = angle
        
        results['%s_flux_fit' % tracer] = xarray.DataArray(fit, dims='along')
        results['%s_flux_fit' % tracer].attrs['emissions'] = q
        results['%s_flux_fit' % tracer].attrs['emissions_std'] = q_std
        results['%s_flux_fit' % tracer].attrs['wind_speed'] = wind_speed
        results['%s_flux_fit' % tracer].attrs['wind_direction'] = wind_direction
        
        if tracer == 'NO2':
            results['%s_flux_fit' % tracer].attrs['decay_time'] = p[1] / wind_speed / 3600
            results['%s_flux_fit' % tracer].attrs['decay_time_std'] = p_std[1] / wind_speed / 3600
            
    
    
    return results
    

def estimate_emissions(source, filename=None, make_figure=True, tracer='NO2',
                       ssp='high', scaling=1.0, redo_preperation=False):
    """
    Estimate emissions of point source.
    """
    origins = sorted(smartcarb.cfg.POINT_SOURCES)
    data, curves = get_mass_balance_data(filename, origins, only_detection=False,
                                         make_figure=False, redo_preperation=redo_preperation,
                                         tracer=tracer, ssp=ssp, scaling=scaling)
    
   
    if source not in data.origin:
        print('"%s" not in swath' % source)
        return None
    
    this = select_source(data, source)
    
    if this['detected_plume'].sum() == 0 or misc.has_multiple_sources(data, source):
        print('No detection or multiple sources for %s' % source)
        return None
    
    if curves.get(source, None) is None:
        print('No curve for %s' % source)
        return None
    
    flux = estimate_fluxes(this, curves[source], origin=source, method='two-gauss')
   
    if flux is None:
        print('Failed to compute any line densities for %s.' % source)
        return None
        
    path = os.path.join('%s-%s-scaling-%.1f' % (tracer, ssp, scaling), source)
    os.makedirs(os.path.join(OUTPUT_PATH, path), exist_ok=True)
    
    flux.attrs.update(
        DESCRIPTION = ('Estimated CO2 and NO2/NOx emissions with ',
                       'mass-balance approach using detected  plumes.'),
        DATAORIGIN = 'SMARTCARB study',
        CREATOR = 'Gerrit Kuhlmann',
        EMAIL = 'gerrit.kuhlmann@empa.ch',
        AFFILIATION = 'Empa Duebendorf, Switzerland',
        DATE_CREATED = pandas.Timestamp.now().strftime('%Y-%m-%d %H:%M'),
    )
    flux_filename = get_output_filename(filename, path=path, extension='_flux.nc')
    flux.to_netcdf(flux_filename)
    
    if make_figure:
        fig = plot_source_overview(data, curves, flux, source, no2_scaling=scaling)
        
        for ext in ['png', 'pdf']:
            figure_fname = get_output_filename(filename, path=path, extension='_flux.%s' % ext)
            fig.savefig(figure_fname)
    
    return flux
   
    
def estimate_emissions_mp(args):
    estimate_emissions(filename=args[0], **args[1])
    plt.close('all')
  
    
def main(source, tracer='NO2', ssp='high', scaling=1.0, n_proc=1):  
    
    with open('/project/kug/repos/smartcarb/plume-detection-algorithm/files/orbits-with-cloud-free-plumes.dat') as file_:
        good_filenames = [line.strip() for line in file_ if not line.startswith('#')]

    good_filenames = (sorted(set(good_filenames)))

    # largest power plants with >3.0 kt NO2 yr-1 emission
    origins = sorted(smartcarb.cfg.POINT_SOURCES)
   
    if n_proc == 1:
        for fn in good_filenames:
            estimate_emissions(source, filename=fn, make_figure=True, tracer=tracer,
                               ssp=ssp, scaling=scaling, redo_preperation=False)
            plt.close()
    else:
        iterator = [(fn, {'source': source, 'make_figure': True, 'tracer': tracer, 'ssp': ssp, 'scaling': scaling})
                    for fn in good_filenames]
        
        with mp.Pool(processes=n_proc) as pool:
            pool.map(estimate_emissions_mp, iterator)

    
    
    